/**
  GPU Flame Rendering Application
  	This version keeps several grid resolutions around
  
  Orion Sky Lawlor, olawlor@acm.org, 2011-10-31 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include <algorithm> /* for std::sort */
#include "osl/mat4.h"
using osl::mat4;
#include "osl/vec4.h" /*<- handy little vector class */
#include "osl/vec2.h"
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z

#include "GL/glui.h" /* user interface */

#include "osl/random_vector.h"
#include "osl/random.cpp"

#include "ogl/image.h" /* (originally from) Nigel Stewart's GLT image library */
#include "ogl/image.cpp"

#include "ogl/glsl.h" /* Orion's tiny GLSL library */
#include "ogl/glsl.cpp"

#include "ogl/minicam.h"
#include "ogl/ppm.h"

/* Debug printing shader code */
std::string verbose_shaders(const char *where,std::string shader) {
	printf("\n\n\n\n-------------- %s ---------------\n%s\n\n\n",where,shader.c_str());
	return shader;
}


int fractal_predefined=1; /* preset fractal */
int animate_maps=0; /* time-varying animation */
int firstlevel=5; /* first multigrid level to compute */
int lastlevel=11; /* last multigrid level to display (==output resolution) */
enum {nlevel=13}; /* highest resolution multigrid texture is 2^nlevel pixels */

int iterations=3; /* per multigrid level */
float screen_initval=1.0, screen_initscale=2.0, map_bias=0.0, map_exposure=1.8, colorful=0.5;
int disc_compute=50; /* (per-vertex) discretization: X and Y divided into this many pieces */
int save_image=0,save_count=1; /* dump rendered image to disk */
int save_movie=0;
float texscale=0.25;
float texscalei=1.0/texscale;
int force_vertexonly=0; /* if true, treat all variations as non-invertable */


/* A flame "variation": a nonlinear function on 2D space. 
Implemented both in C++ (mostly for testing/debugging)
and in GLSL via the macros below.
*/
class variation {
public:
	/* destructor */
	virtual ~variation();
	
	/* Map this location in 2D space */
	virtual vec2 apply(const vec2 &t)=0;
	
	/* Return offical human-readable name of this variation */
	virtual std::string getName(void) const =0;
	
	/* Return GLSL-compatible code that reads a vec2 "t" 
	   and writes to the vec2 "p", applying this variation. */
	virtual std::string getCode(void)=0;
	
	
	
	/* Return true if this function has an (easy-to-compute algebraic) inverse */
	virtual bool hasInverse(void) {return false;}
	
	/* Call the function f on each of the inverse images 
	   of this point under this function. */
	typedef vec4 (*fn_taking_vec2_t)(const vec2 &t);
	virtual vec4 sumInverse(const vec2 &p,fn_taking_vec2_t f)
		{return vec4(0.0);}
	
	/* Return GLSL code that calls the function "f" with 
	   the vec2 "t" for each inverse of the vec2 "p", and
	   returns the vec4 sum of all the calls to "f".
	*/
	virtual std::string getInverseCode(void) {return "";}
	
	
	/* Return the Jacobian determinant at this point */
	virtual float getArea(const vec2 &t)=0;
	
	/* Return the GLSL code that returns the float jacobian
	determinant, given the float source location (x,y). */
	virtual std::string getAreaCode(void)=0;
};

variation::~variation() {}

/* Linearly combine two nonlinear variations.  Normally non-invertable. 
This class would be useful for animations, but it's not used right now.
*/
class variationpair : public variation {
public:
	variation *a,*b;
	float wa,wb;
	variationpair(variation *a_,float wa_,
		variation *b_,float wb_)
		:a(a_),b(b_),wa(wa_),wb(wb_) {}
	virtual vec2 apply(const vec2 &src) {
		return wa*a->apply(src)+wb*b->apply(src);
	}
	virtual std::string getName(void) const {
		char nameStr[100];
		sprintf(nameStr,"(%f*%s+%f*%s)",
			wa,a->getName().c_str(),
			wb,b->getName().c_str());
		return nameStr;
	}
	virtual std::string getCode(void) {
		char weightStr[100];
		sprintf(weightStr,"%f*pa+%f*p\n;\n",wa,wb);
		return "vec2 pa;\n{" +
			a->getCode() +
			"} pa=p; {\n" +
			b->getCode() +
			"} p="+std::string(weightStr);
	}
};


/*
  FUNKY macros to generate both C++ and GLSL versions of each variation.
  The code is passed in as a macro argument, and either:
  	substituted in verbatim (for C++)
  or	stringified with # and then returned (for GLSL)
  
  The big advantage is I can cut down on the amount of duplicated code per variation.
*/
#define MAKE_VARIATION(name,code,jacobian) \
class variation_##name : public variation { /* class */\
public:\
	virtual vec2 apply(const vec2 &t) {\
		vec2 p;\
		{ code } \
		return p;\
	} \
	virtual std::string getName(void) const { \
		return #name; \
	} \
	virtual std::string getCode(void) { \
		return #code; \
	}\
	virtual float getArea(const vec2 &t) {\
		jacobian \
	} \
	virtual std::string getAreaCode(void) {\
		return #jacobian; \
	}\
};  \
variation_##name singleton_##name; /* standard instance */

#define MAKE_VARIATION_INVERSE(name,code,icode,jacobian) \
class variation_##name : public variation { /* class */\
public:\
	virtual vec2 apply(const vec2 &t) {\
		vec2 p;\
		{ code } \
		return p;\
	} \
	virtual std::string getName(void) const { \
		return #name; \
	} \
	virtual std::string getCode(void) { \
		return #code; \
	}\
	virtual bool hasInverse(void) {return true;} \
	virtual vec4 sumInverse(const vec2 &p,fn_taking_vec2_t f) {\
		icode \
	} \
	virtual std::string getInverseCode(void) {\
		return #icode;\
	}\
	virtual float getArea(const vec2 &t) {\
		jacobian \
	} \
	virtual std::string getAreaCode(void) {\
		return #jacobian; \
	}\
};  \
variation_##name singleton_##name; /* standard instance */


// C++ versions of standard constants.  These are in math.h on Linux.
//const float M_PI=3.14159265358979323;
//const float M_1_PI=1.0/M_PI;
#ifndef M_PI 
#define M_PI 3.14159265358979
#endif
#ifndef M_1_PI
#define M_1_PI (1.0/M_PI)
#endif
const float EPS=1.0e-6;

MAKE_VARIATION_INVERSE(linear,
	p=t;
,/*inverse*/
	return f(p);
,/* jacobian */
	return 1.0;
);
MAKE_VARIATION_INVERSE(sinusoidal,
	p=vec2(sin(t.x),sin(t.y));
,/*inverse*/
	if (abs(p.x)>1.0 || abs(p.y)>1.0) return vec4(0.0); /* out of bounds */
	vec4 sum=vec4(0.0);
	vec2 t=vec2(asin(p.x),asin(p.y));
	float img=2; // <- 0-2 is reasonable.  theoretically should be infinity!
	for (float dy=-img;dy<=img;dy++)
	for (float dx=-img;dx<=img;dx++) 
	{
		sum+=f(vec2( t.x+M_PI*2.0*dx,t.y+M_PI*2.0*dy));
		sum+=f(vec2(-t.x+M_PI*(1.0+2.0*dx),t.y+M_PI*2.0*dy));
		sum+=f(vec2(-t.x+M_PI*(1.0+2.0*dx),-t.y+M_PI*(1.0+2.0*dy)));
		sum+=f(vec2( t.x+M_PI*2.0*dx,-t.y+M_PI*(1.0+2.0*dy)));
	}
	return sum;
,/*jacobian*/
	return cos(t.x)*cos(t.y);
);
MAKE_VARIATION_INVERSE(spherical,
	float r2=dot(t,t) + EPS;
	p=t/r2;
,/*inverse*/
	float r2=dot(p,p);
	return f(p/r2);
,/*jacobian*/
	float r2=dot(t,t);
	return 1.0/(r2*r2);
);
MAKE_VARIATION_INVERSE(swirl,
	float r2=dot(t,t);
	float s=sin(r2); float c=cos(r2);
	p=vec2(
		s*t.x-c*t.y,
		c*t.x+s*t.y
	);
,/*inverse*/
	float r2=dot(p,p);
	if (r2>10.0) return vec4(0.0); /* give up--mipmap can't handle shear... */
	float s=sin(r2); float c=cos(r2);
	return f(vec2(
		s*p.x+c*p.y,
		-c*p.x+s*p.y
	));
,/*jacobian*/
	return 1.0;
);
MAKE_VARIATION_INVERSE(horseshoe,
	float r=1.0/(length(t)+EPS);
	p=vec2(
		(t.x-t.y)*(t.x+t.y)*r,
		2.0*t.x*t.y*r
	);
,/*inverse: slight numerical problems for big x and small y*/
	float r2 = p.x*p.x + p.y*p.y;
	float r=sqrt(r2);
	float aspect = (p.x + r)/p.y;
	float y = sqrt(0.5*(r2 - p.x*r));
	return f(vec2(y*aspect,y))
	      +f(vec2(-y*aspect,-y));
,/*jacobian*/
	return 2.0;
);
MAKE_VARIATION_INVERSE(polar,
	p=vec2(
		atan2(t.x,t.y)*M_1_PI,
		length(t)-1.0
	);
,/*inverse*/
	if (p.x>=-1.0 && p.x<=1.0 && p.y>-1.0)
		return f((p.y+1)*vec2(sin(p.x*M_PI),cos(p.x*M_PI)));
	else
		return vec4(0.0);
,/*jacobian*/
	return M_1_PI/length(t);

);
MAKE_VARIATION(handkerchief,
	float a=atan2(t.x,t.y);
	float r=length(t);
	p=r*vec2(
		sin(a+r),
		cos(a-r)
	);
,/* jacobian */
	float r=length(t);
	return cos(2.0*r) + 2.0*t.x*t.y/r - r*sin(2.0*r);
);
MAKE_VARIATION_INVERSE(heart,
	float a=atan2(t.x,t.y);
	float r=length(t);
	//printf("H(%.2f,%.2f -> %.2f,%.2f) ",t.x,t.y,a,r);
	p=vec2(
		r*sin(r*a),
		(-r)*cos(r*a)
	);
,/* inverse */
	float r=length(p);
	float at2=atan2(p.x,-p.y);
	
	/* Figure out the correct multiple(s) of 2PI to add
	 to our arctangent above: */
	float pilo=ceil((-M_PI*r-at2)/(2.0*M_PI));
	float pihi=floor((+M_PI*r-at2)/(2.0*M_PI));
	vec4 sum=vec4(0.0);
	for (float pic=pilo;pic<=pihi;pic++) {
		float a=(at2+pic*2.0*M_PI)/r;
		//printf("HI(%.1f,%.2f,%.2f) ",pic,a,r);
		sum+=f(r*vec2(sin(a),cos(a)));
	}
	return sum;
,/* jacobian */
	return length(t);
);
MAKE_VARIATION(disc,
	float a=atan2(t.x,t.y)*M_1_PI;
	float r=length(t)*M_PI;
	p=a*vec2(
		sin(r),
		cos(r)
	);
/*(self) inverse: This is *only* valid for input points in the 1st quadrant with r<1.0! 
	float a=length(p)*M_PI;
	float r=atan2(p.x,p.y)*M_1_PI;
	return f(vec2(r*sin(a),r*cos(a)));*/
,/* jacobian */
	float a=atan2(t.x,t.y);
	return M_1_PI*a/length(t);
);
MAKE_VARIATION(spiral,
	float a=atan2(t.x,t.y);
	float r=length(t)+EPS;
	float r1=1.0/r;
	p=r1*vec2(
		cos(a)+sin(r),
		sin(a)-cos(r)
	);
,/* jacobian */
	float r2=dot(t,t);
	float r=sqrt(r2);
	float a=atan2(t.x,t.y);
	return (1 - r*cos(r - a) + sin(r - a))/r2;
);
MAKE_VARIATION_INVERSE(hyperbolic,
	/* Was:
	float a=atan2(t.x,t.y);
	float r=length(t) + 1.0e-6;
	p=vec2(sin(a)/r,cos(a)*r);
	*/
	float r2=dot(t,t)+1.0e-6;
	p=vec2(t.x/r2,t.y); /* simplified w/trig identities */
,/*inverse*/
	float v01 = 1.0/p.x;
	float det=1.0 - 4.0*p.x*p.x*p.y*p.y;
	if (det>=0.0) {
		float v03 = sqrt(det);
		return f(vec2(((v01*(1.0 - v03))/2.0),p.y))
		      +f(vec2(((v01*(1.0 + v03))/2.0),p.y));
	} else { /* negative determinant--no solutions */
		return vec4(0.0);
	}

,/* jacobian */
	/* Was:
		float a=atan2(x,y);
		return cos(2*a)/r2;
	
	I'm sure there's some way to simplify this, e.g.,
	with cos(2*a) = 2 (cos a)^2-1
	cos(a)*r=y; cos(a)=y/r; 
	cos(2a)=2 y^2/r^2-1
	*/
	float r2=dot(t,t);
	return (1.0-2.0*t.y*t.y/r2)/r2;
);
MAKE_VARIATION(diamond,  /* has an inverse, but like 32 roots! */
	float a=atan2(t.x,t.y);
	float r=length(t);
	p=vec2(sin(a)*cos(r),cos(a)*sin(r));
,/* jacobian */
	float r2=dot(t,t);
	float r=sqrt(r2);
	return (cos(2.0*r)+2*t.y*t.y/r2-1)/(2.0*r);
);
MAKE_VARIATION(ex,
	float a=atan2(t.x,t.y);
	float r=length(t);
	float n0=sin(a+r);
	float n1=cos(a-r);
	float m0=n0*n0*n0*r;
	float m1=n1*n1*n1*r;
	p=vec2(m0+m1,m0-m1);
,/* jacobian */
	float r2=dot(t,t);
	float r=sqrt(r2);
	float s2r=sin(2.0*r);
	float c2r=cos(2.0*r);
	float inner=s2r+t.x*t.y/r2;
	return 1.5/r*(6.0*t.x*t.y+r*c2r-3*r2*s2r)*inner*inner;
);

/* FIXME: julia is random, but only two possible outcomes--
  so split incoming maps, make other map have negative wval 
  to flip output around */
MAKE_VARIATION_INVERSE(julia,
	/* Trig version:
	float a=atan2(t.x,t.y);
	float r=length(t);
	p=sqrt(r)*vec2(cos(a*0.5),sin(a*0.5));
	*/
	
	// Man, trig identities are friggin' weird.  Half-angle gives:
	float r=length(t);
	p=vec2(sqrt(0.5*(r+t.y)),sqrt(0.5*(r-t.y)));
	if (t.x<0.0) p.y=-p.y; /* flip sign */
	
,/* inverse */
	if (p.x<0) return vec4(0.0); /* that point's not one of ours */
	/* Triggy inverse is this:
	float a=atan2(p.y,p.x);
	float r2=dot(p,p);
	return f(r2*vec2(sin(a*2.0),cos(a*2.0)));
	*/
	// Mathematica claims this is also the inverse:
	return f(vec2( 2*p.x*p.y, (p.x - p.y)*(p.x + p.y) ));
	
,/* jacobian */
	return 1.0/(4.0*length(t));
);

MAKE_VARIATION_INVERSE(bent,
	vec2 n=t;
	if (n.x<0.0) n.x*=2.0;
	if (n.y<0.0) n.y*=0.5;
	p=n;
,/* inverse */
	vec2 n=p;
	if (n.x<0.0) n.x*=0.5;
	if (n.y<0.0) n.y*=2.0;
	return f(n);
,/* jacobian */
	float scale=1.0;
	if (t.x<0.0) scale=2.0;
	if (t.y<0.0) scale*=0.5;
	return scale;
);
/* FIXME: waves somehow adjusts parameters based on coeffs */
MAKE_VARIATION_INVERSE(fisheye,
	float r=length(t);
	r=2/(r+1);
	p=r*vec2(t.y,t.x); //<- note backwards!
	/* length (p) = s
		s=r*2/(r+1)  (w/ extra copy of length(t) from vec2)
		sr+s=r*2   (solve for r)
		r*(s-2)=-s
		r=s/(2-s)
	*/
,/*inverse*/
	float s=length(p);
	float r=1.0/(2.0-s);
	if (r>=0.0) /* plausible fisheye output: */
		return f(r*vec2(p.y,p.x));
	else /* outside fisheye region */
		return vec4(0.0);
,/*jacobian*/
	float r=length(t);
	float r1=1.0+r;
	return 4.0/(r1*r1*r1);
);
/* FIXME: popcorn needs coeffs too */
MAKE_VARIATION_INVERSE(exponential, 
	float dx=exp(t.x-1.0);
	float dy=M_PI*t.y;
	p=dx*vec2(cos(dy),sin(dy));
,/*inverse*/
	float r=length(p);
	float a=atan2(p.y,p.x);
	return f(vec2(log(r)+1.0,a*M_1_PI));
,/*jacobian*/
	return M_PI*exp(2.0*t.x-2.0);
);
MAKE_VARIATION_INVERSE(power,
	float a=atan2(t.x,t.y);
	p=vec2(cos(a),sin(a))*pow(length(t),sin(a));
,/*inverse*/
	float a=atan2(p.y,p.x);
	float r=pow((float)length(p),(float)1.0/sin(a));
	if (r==0.0 || r+1.0==r) /* r has exploded--give up */
		return vec4(0.0);
	else /* returned a reasonable value */
		return f(r*vec2(sin(a),cos(a)));
,/*jacobian*/
	float ir=1.0/length(t);
	float r=1.0/ir;
	return pow((float)r,(float)(2.0*(t.x*ir)-2.0))*(t.x*ir);
);
MAKE_VARIATION(cosine,  /* has 16 families of solutions */
	float n=t.x*M_PI;
	p=vec2(cos(n)*cosh(t.y),-sin(n)*sinh(t.y));
,/*jacobian*/
	return 0.5*M_PI*(-cos(2.0*M_PI*t.x)+cosh(2.0*t.y));
);
/* FIXME: rings needs coeffs */

variation *variation_list[]={
	&singleton_linear,
	&singleton_sinusoidal,
	&singleton_spherical,
	&singleton_swirl,
	&singleton_horseshoe,
	&singleton_polar,
	&singleton_handkerchief,
	&singleton_heart,
	&singleton_disc,
	&singleton_spiral,
	&singleton_hyperbolic,
	&singleton_diamond,
	&singleton_ex,
	&singleton_julia,
	&singleton_bent,
	&singleton_fisheye,
	&singleton_exponential,
	&singleton_power,
	&singleton_cosine,
	0
};

int variation_lookup(const variation *v) {
	for (int i=0;variation_list[i];i++)
		if (v==variation_list[i])
			return i;
	printf("ERROR: Could not locate variation %s!\n",v->getName().c_str());
	return -1;
}



#if 0 /* Do inverse and jacobian code verification at runtime */
/* Silly inverse verification testing/printing code: */
variation *curvar;
vec2 p;
vec4 f(const vec2 &t) {
	vec2 hp=curvar->apply(t);
	if (length(hp-p)>0.01) {
		printf("    bad %s inverse at point %f,%f  (hp=%f,%f, t=%f,%f)\n",
			curvar->getName().c_str(),
			p.x,p.y,
			hp.x,hp.y,
			t.x,t.y
		);
		//abort();
	}
	return vec4(1.001234);
}

/* Verify that this variation's inverse actually works: 
that I can start with random plane points, invert them,
and each of the inverted images maps back to the original point.
*/
void verify_inverse(variation *var,int seed) {
	curvar=var;
	osl::Random48 r(seed);
	static int badcount=0;
	bool isbad=false;
	for (int rep=0;rep<10000;rep++) {
		p=vec2(20.0*r.nextFloat()-10.0,
			20.0*r.nextFloat()-10.0);
		if (var==&singleton_disc) /* *very* limited range for inverse */
			p=vec2(r.nextFloat()*0.7,r.nextFloat()*0.7);
		//Check inverse:
		vec4 v=var->sumInverse(p,f);
		
		//Check jacobian:
		float e=1.0e-4;
		vec2 t=p*0.1; /* limit range, to reduce roundoff */
		vec2 p0=var->apply(t);
		vec2 p1=var->apply(t+vec2(e,0));
		vec2 p2=var->apply(t+vec2(0,e));
		float area=fabs(p0.ccw(p1,p2)/(e*e));
		float jacobi=fabs(var->getArea(t));
		if (fabs((area-jacobi)/area)>0.1 &&
		    fabs(area-jacobi) > 1.0e-2) 
		{ /* Mismatch detected */
			printf("%d:    bad %s jacobian at point %f,%f  (area=%f, jacobi=%f)\n",
				badcount++,curvar->getName().c_str(),
				t.x,t.y,
				area,jacobi
			);
			isbad=true;
		}
		//else if (isbad) printf("."); /* indicating good jacobian */
	}
}

#else /* no verification */
void verify_inverse(variation *var,int seed) {/* ignore it */}
#endif

/* Verify that all defined variation inverses actually work. */
void verify_inverses(void) {
	for (unsigned int i=0;i<sizeof(variation_list)/sizeof(variation_list[0]);i++)
	{
		if (variation_list[i] && variation_list[i]->hasInverse())
			verify_inverse(variation_list[i],i);
	}
}


/* STOOPID hack for C++-side xForm inverse testing */
variation::fn_taking_vec2_t xform_f=0;
vec2 xform_x,xform_y,xform_o;
vec4 xform_f_fn(const vec2 &t) {
	vec2 t2=xform_x*t.x+xform_y*t.y+xform_o;
	return xform_f(t2);
}

/* One map from a set of flame transforms */
class xForm : public variation {
public:
	float weight; /* this transform's weight */
	vec2 x,y,o; /* linear part of transform */
	float wvar; /* nonlinear weight */
	//variation *var; /* nonlinear portion of transform */
	int var; /* index into variation_list */
	std::string tag; /* GLSL tag to append to function names */
	
	// used to verify cached xForm parameters
	bool operator==(const xForm &b) const {
		//if (weight!=b.weight || x!=b.x || y!=b.y || o!=b.o || wvar!=b.wvar) // <- now in uniforms!
		if (var!=b.var)
			return false;
		else return true;
	}
	
	// Used to initialize from draves-style file
	void init(float w,variation *var_ptr,const char *coef_str) {
		static int count=0;
		char buf[100]; sprintf(buf,"%d",count++); tag=buf;
		weight=w;
		sscanf(coef_str,"%f %f %f %f %f %f",
			&x.x,&x.y,&y.x,&y.y,&o.x,&o.y);
		wvar=1.0;
		var=variation_lookup(var_ptr);
	}
	void init(float w,variation *var_ptr,float v,const char *coef_str) {
		init(w,var_ptr,coef_str); wvar=v;
	}
	
	xForm(std::string tag_="") 
		:weight(0.0),
		 x(1.0,0.0),y(0.0,1.0),o(0.0,0.0),
		 wvar(1.0),var(0),tag(tag_) {}
	
	/* Map this location in 2D space */
	virtual vec2 apply(const vec2 &src)
	{
		vec2 t=x*src.x+y*src.y+o;
		return wvar*variation_list[var]->apply(t);
	}
	
	/* Return offical human-readable name of this variation */
	virtual std::string getName(void) const {
		return "xForm"+tag;
	}
	
	/* Return GLSL-compatible code that reads a vec2 "t" 
	   and writes to the vec2 "p". */
	virtual std::string getCode(void) {
		/* SILLY: xForm needs a top-level function wrapper (should push this outside somewhere) */
		std::string ret=getUniformDecl()+
			"vec2 applyMap"+tag+"(vec2 t) {\n"
			"	t=xf"+tag+"[0]*t.x+xf"+tag+"[1]*t.y+xf"+tag+"[2];\n"
			" 	vec2 p;\n"
			"{	"+variation_list[var]->getCode()+"\n}\n"
			"	return xf"+tag+"[6].x*p;\n"
			"}\n"
			"float jacobian"+tag+"(vec2 t) {\n"
				+variation_list[var]->getAreaCode()+
			"}\n"
			"float getDensity(vec2 t) {\n"
			"	t=xf"+tag+"[0]*t.x+xf"+tag+"[1]*t.y+xf"+tag+"[2];\n"
			"	return 1.0/(1.0e-3+abs(xf"+tag+"[7].x*jacobian"+tag+"(t)));\n"
			"}\n";
		std::cout<<"Shader map GLSL code: "<<ret<<"\n\n\n";
		return ret;
	}
	
	/* Return true if this function has an (easy-to-compute algebraic) inverse */
	virtual bool hasInverse(void) {
		if (tag=="NOINV") return false; // FIXME: hack to show how rasterization works!
		if (force_vertexonly) return false;
		return variation_list[var]->hasInverse();
	}
	
	/* Call the function f on each of the inverse images 
	   of this point under this function. */
	virtual vec4 sumInverse(const vec2 &p,fn_taking_vec2_t f)
	{ 
		/*STOOPID hack: call variation's sumInverse, with our f, in turn real f */
		xform_f=f;
		makeInverseMatrix(xform_x,xform_y,xform_o);
		vec2 p2=(1.0/wvar)*p;
		return variation_list[var]->sumInverse(p2,xform_f_fn);
	}
	
	/* Return GLSL code that calls the function f for each inverse. 
	
	Code chunks involved:
		g: plane coords to texture samples (with main, outside)
		f: invert linear transform, call g
		sumInverse: call f at each nonlinear inverse
		main: call sumInverse at plane coords (also outside)
	*/
	virtual std::string getInverseCode(void) {
		return  getUniformDecl()+
			"float jacobian"+tag+"(vec2 t) {"
			+variation_list[var]->getAreaCode()+
			"}\n"
			"vec4 f"+tag+"(vec2 inv) { /* called by nonlinear inverters, invert linear part of map */ \n"
			"	float areaScale=1.0/(1.0e-2+abs(xf"+tag+"[7].x*jacobian"+tag+"(inv)));\n"
			"	vec2 p=xf"+tag+"[3]*inv.x+xf"+tag+"[4]*inv.y+xf"+tag+"[5];\n"
			"	return areaScale*g"+tag+"(p);\n"
			"}\n"
			"vec4 nonlinear_inverse"+tag+"(vec2 p) {\n"
			"	p=p*xf"+tag+"[6].y;\n"
			"	"+sed(variation_list[var]->getInverseCode(),"f(","f"+tag+"(")+"\n"
			"}\n";
	}
	
	/* Return the affine (non-variation) component of our jacobian determinant */
	float getAffineArea(void) const {
		return fabs(wvar*wvar*(x.x*y.y-x.y*y.x));
	}
	
	// Number of GLSL uniform vec2's in array below
	enum {nUniforms=8};

	/* Fill this vector with our uniform variables "xf"+tag */
	void getUniforms(vec2 *xf) const {
		xf[0]=x; xf[1]=y; xf[2]=o;
		makeInverseMatrix(xf[3],xf[4],xf[5]);
		xf[6]=vec2(wvar,1.0/wvar);
		xf[7]=vec2(getAffineArea(),0.0);
	}
	/* Apply our uniforms to this shader */
	void setUniforms(GLhandleARB prog) const {
		vec2 xf[nUniforms];
		getUniforms(xf);
		float *f=xf[0];
		glUniform2fvARB(glGetUniformLocationARB(prog,
			("xf"+tag).c_str()),nUniforms,f);
	}
	
	/* Return the Jacobian determinant at this point */
	virtual float getArea(const vec2 &src) {
		vec2 t=x*src.x+y*src.y+o;
		return variation_list[var]->getArea(t)*getAffineArea();
	}
	
	/* Return the GLSL code that returns the float jacobian
	determinant, given the float source location (x,y). */
	virtual std::string getAreaCode(void) {
		return "NOOOO!!!!!  xForm::getAreaCode not implemented";
	}
private:
	std::string getUniformDecl(void) const {
		return "uniform vec2 xf"+tag+"[8]; // Forward xyo[0-2], inverse XYO[3-5], vec2(wvar,1.0/wvar)[6], affinearea[7]\n";
	}

	/* Make X, Y, and O columns of the inverse of our linear matrix x,y,o
	
	Goal:
		b=x*a.x+y*a.y+o; // forward transform, in apply
		a=X*b.x+Y*b.y+O; // inverse transform
	
	*/
	void makeInverseMatrix(vec2 &X,vec2 &Y,vec2 &O) const {
		float idet=1.0f/(x.x*y.y-x.y*y.x);
		X.x=idet*y.y;  Y.x=idet*-y.x;
		X.y=idet*-x.y; Y.y=idet*x.x;
		O=-(X*o.x+Y*o.y); /* see a==0 case */
	}
	
	/* Utility function: substitute all occurrences of A with B in src. */
	std::string sed(const std::string &src,const std::string A,const std::string B) {
		std::string ret;
		size_t lastpos=0, srcpos=0;
		while (std::string::npos!=(srcpos=src.find(A,lastpos))) {
			//printf("xForm::sed> A=%d, last=%d, ret='%s'\n",srcpos,lastpos,ret.c_str());
			/* Found a copy of A at position srcpos */
			ret+=src.substr(lastpos,srcpos-lastpos)+B;
			srcpos+=A.size(); /* skip over entire A string */
			lastpos=srcpos;
		}
		return ret+src.substr(lastpos); // tail of string unmodified
	}
};

/* Hardcode a single flame, for testing purposes */
std::vector<xForm> make_xfms(void) {
	std::vector<xForm> xfm;
	xForm xf;
	switch(fractal_predefined) {
	case -1: // Standard benchmark
		screen_initval=0.1;
		screen_initscale=66;
		map_exposure=1.5;
		colorful=0.6;
		texscale=0.25;
		
		xf.tag=("0");
		xf.weight=0.377;
		xf.x=vec2(-0.360294,-0.439868);
		xf.y=vec2(0.94579,-0.861994);
		xf.o=vec2(1.120671,-1.359398);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_hyperbolic);
		xfm.push_back(xf);

		xf.tag=("1");
		xf.weight=0.298;
		xf.x=vec2(0.871566,0.250037);
		xf.y=vec2(0.467305, 0.979817 );
		xf.o=vec2(-0.990552, -0.538029);
		xf.wvar=1;
		//xf.wvar=0.8; 
		xf.var=variation_lookup(&singleton_linear);
		xfm.push_back(xf);

		xf.tag=("2");
		xf.weight=0.325;
		xf.x=vec2(-0.326387, 0.395477 );
		xf.y=vec2(-0.415395, -0.380892 );
		xf.o=vec2(-0.782861, -1.30803);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_hyperbolic); // ex; //<- has no inverse
		xfm.push_back(xf);
		break;
	case -2: // Caberet
		screen_initval=0.4;
		screen_initscale=66;
		map_exposure=-1.6;
		colorful=0.55;
		texscale=0.25;
		
		xf.tag=("0");
		xf.weight=0.4;
		xf.x=vec2(1,-1);
		xf.y=vec2(0.5,0.5);
		xf.o=vec2(-0.3,-2.1);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_polar);
		xfm.push_back(xf);

		xf.tag=("1");
		xf.weight=0.3;
		xf.x=vec2(1,0);
		xf.y=vec2(0,1);
		xf.o=vec2(-1.2,-0.9);
		xf.wvar=1;
		//xf.wvar=0.8; 
		xf.var=variation_lookup(&singleton_linear);
		xfm.push_back(xf);

		xf.tag=("2");
		xf.weight=0.4;
		xf.x=vec2(-0.2, 0.4 );
		xf.y=vec2(-0.4, -0.3);
		xf.o=vec2(-1.0, -1.3);
		xf.wvar=1.65; xf.var=variation_lookup(&singleton_spherical);
		xfm.push_back(xf);
		break;
	case 0: // Fountains
		screen_initval=0.4;
		screen_initscale=66;
		map_exposure=-1.6;
		colorful=0.55;
		texscale=0.25;
		
		xf.tag=("0");
		xf.weight=0.4;
		xf.x=vec2(1,-1);
		xf.y=vec2(0.5,0.5);
		xf.o=vec2(0.6,-3.0);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_polar);
		xfm.push_back(xf);

		xf.tag=("1");
		xf.weight=0.3;
		xf.x=vec2(1,0);
		xf.y=vec2(0,1);
		xf.o=vec2(-1.995,-0.0);
		xf.wvar=1;
		//xf.wvar=0.8; 
		xf.var=variation_lookup(&singleton_linear);
		xfm.push_back(xf);

		xf.tag=("2");
		xf.weight=0.3;
		xf.x=vec2(-0.2, 0.4 );
		xf.y=vec2(-0.4, -0.3);
		xf.o=vec2(-1.0, -1.3);
		xf.wvar=1.8; xf.var=variation_lookup(&singleton_spherical);
		xfm.push_back(xf);
		break;
	case 1: // *Pretty* two-map spiral
		screen_initval=0.1;
		screen_initscale=66;
		map_exposure=-1.5;
		colorful=0.3;
		texscale=0.25;

		xf.tag=("0");
		xf.weight=0.5;
		xf.x=vec2(1,0);
		xf.y=vec2(0,1);
		xf.o=vec2(0,0);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_hyperbolic);
		xfm.push_back(xf);

		xf.tag=("1");
		xf.weight=0.5;
		xf.x=vec2(0.8,0.4);
		xf.y=vec2(-0.4,0.8);
		xf.o=vec2(-1,0.3);
		//xf.wvar=1;
		xf.wvar=1.0; 
		xf.var=variation_lookup(&singleton_linear); // spherical;
		xfm.push_back(xf);
		break;
	case 2: // "Eyes": Hyperbolic, fisheye, and linear
		screen_initval=0.1;
		screen_initscale=66;
		map_exposure=-1.5;
		colorful=0.3;
		texscale=0.25;

		xf.tag=("0");
		xf.weight=0.5;
		xf.x=vec2(1,0);
		xf.y=vec2(0,1);
		xf.o=vec2(0,0);
		xf.wvar=0.5; xf.var=variation_lookup(&singleton_fisheye);
		xfm.push_back(xf);

		xf.tag=("1");
		xf.weight=0.4;
		xf.x=vec2(0.8,0.4);
		xf.y=vec2(-0.4,0.8);
		xf.o=vec2(-1,0.3);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_linear); // spherical;
		xfm.push_back(xf);
		
		xf.tag=("2");
		xf.weight=0.1;
		xf.x=vec2(1,1);
		xf.y=vec2(-1,1);
		xf.o=vec2(0,0);
		xf.wvar=3.2; xf.var=variation_lookup(&singleton_hyperbolic); // spherical;
		xfm.push_back(xf);
		break;
	case 3: /* Swirlpinski */
		firstlevel=8; lastlevel=11; // swirl hits mipmap troubles at low res
		iterations=5;
		screen_initval=1.0; screen_initscale=2.0; map_bias=0.0; map_exposure=1.8; colorful=0.5;
		texscale=0.25;
	{
		float scale=1.3; /* controls amount of nonlinearity experienced */
		xf.tag=("0");
		xf.weight=0.333;
		xf.x=vec2(0,0.5);
		xf.y=vec2(0.5,0);
		xf.o=vec2(0,0);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_swirl);
		xfm.push_back(xf);

		xf.tag=("1");
		xf.weight=0.333;
		xf.x=vec2(0.5,0);
		xf.y=vec2(0,0.5);
		xf.o=vec2(+scale,0);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_linear); // spherical;
		xfm.push_back(xf);

		xf.tag=("2");
		xf.weight=0.333;
		xf.x=vec2(0.5,0);
		xf.y=vec2(0,0.5);
		xf.o=vec2(0.3*scale,-scale);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_linear); // spherical;
		xfm.push_back(xf);
	}
		break;
	case 4: /* unbounded affine */
		screen_initval=1.0; screen_initscale=1.0; map_exposure=1.35; colorful=0.0;
		texscale=0.82; texscalei=1.0/texscale;
		xf.tag=("0");
		xf.weight=0.5;
		xf.x=vec2(1.5,0);
		xf.y=vec2(0,0.2);
		xf.o=vec2(0.25,-0.4);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_linear);
		xfm.push_back(xf);
		
		xf.tag=("1");
		xf.weight=0.5;
		xf.x=vec2(0,0.8);
		xf.y=vec2(0.8,0.0);
		xf.o=vec2(0,0);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_linear);
		xfm.push_back(xf);
		break;
	case 5: // Demo of bad vertex rendering
		texscale=0.5; texscalei=1.0/texscale;
		screen_initval=0.1;
		screen_initscale=66;
		map_exposure=-1.0;
		colorful=0.4;
		//firstlevel=lastlevel=11;
		//iterations=10;

		xf.tag=("0");
		xf.weight=0.33;
		xf.x=vec2(-0.1,1);
		xf.y=vec2(1,0.1);
		xf.o=vec2(0.32,-1);
		xf.wvar=1.0; xf.var=variation_lookup(&singleton_hyperbolic);
		xfm.push_back(xf);

		xf.tag=("1"); // "NOINV"
		xf.weight=1.3;
		xf.x=vec2(1,-0.3);
		xf.y=vec2(0.3,1);
		xf.o=vec2(0,0);
		xf.wvar=2;
		xf.var=variation_lookup(&singleton_sinusoidal);
		xfm.push_back(xf);
		break;
	case 6: // Draves Gen165 Sheep: 
		// Gen 165 sheep 48  (Nice!)
		map_exposure=2.0; colorful=0.5;
		xf.init(0.3333,&singleton_spherical,0.4,"0.5 0 0 0.5 -0.566 0.400"); xfm.push_back(xf);
		xf.init(0.3333,&singleton_spherical,0.4,"0.5 0 0 0.5 0.566 0.400"); xfm.push_back(xf);
		xf.init(0.3333,&singleton_spherical,0.4,"0.5 0 0 0.5 0.000 -0.551"); xfm.push_back(xf);
   
   	/* 	// Gen 165 sheep 0:
		xf.init(0.3333,&singleton_hyperbolic,"0.692904 -0.898884 -0.862574 -0.799124 -0.535423 -0.99203"); xfm.push_back(xf);
		xf.init(0.3333,&singleton_hyperbolic,"0.550665 -0.0437077 -0.44848 -0.613512 0.147667 -0.483213"); xfm.push_back(xf);
		xf.init(0.3333,&singleton_hyperbolic,"-0.282792 -0.0685712 0.507576 -0.750728 0.236344 -0.0978886"); xfm.push_back(xf);
	*/
	/*
		// Gen 165 sheep 600, actually got a vote, looks OK (but per-vertex)
		xf.init(0.25,&singleton_handkerchief,"-0.0800081 -0.741921 -0.287797 -0.593438 -0.10147 0.079855"); xfm.push_back(xf);
		xf.init(0.25,&singleton_handkerchief,"-0.146646 0.337478 0.382939 -0.854078 -0.119577 0.770815"); xfm.push_back(xf);
		xf.init(0.25,&singleton_handkerchief,"-0.36433 0.837788 -0.614212 -0.9201 0.788369 0.325727"); xfm.push_back(xf);
		xf.init(0.25,&singleton_handkerchief,"0.937976 -0.992211 0.25629 0.957151 -0.956824 0.798349"); xfm.push_back(xf);
	*/
	
	/* Stinks:
		xf.init(0.3333,&singleton_fisheye,"-0.970145 -0.984873 0.26221 0.946498 -0.606021 0.473919"); xfm.push_back(xf);
		xf.init(0.3333,&singleton_fisheye,"0.929834 0.657806 0.923901 0.361493 -0.997904 -0.16405"); xfm.push_back(xf);
		xf.init(0.3333,&singleton_fisheye,"0.235448 -0.599886 0.571187 0.322248 -0.871327 0.469188"); xfm.push_back(xf);
	
		xf.init(0.3333,&singleton_horseshoe,"0.521936 0.828349 0.983978 -0.841927 0.701249 -0.873424");xfm.push_back(xf);
		xf.init(0.3333,&singleton_horseshoe,"0.822149 0.264719 -0.384574 0.590485 -0.5294 0.181069");xfm.push_back(xf);
		xf.init(0.3333,&singleton_horseshoe,"-0.726014 -0.934578 0.799966 0.127155 -0.179398 0.991218");xfm.push_back(xf);
	*/
		break;
	};

	return xfm;
}

void oglExit(const char *where,const char *why) {
	printf("Fatal error> %s\n",why);
	exit(1);
}
void bad_framebuffer(GLenum status) {
        switch(status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
                oglExit("oglFramebuffer::bind","framebuffer is actually OK (?!)");
                break;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
                oglExit("oglFramebuffer::bind","combination of formats is UNSUPPORTED by your card");
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
                oglExit("oglFramebuffer::bind","ATTACHMENT_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
                oglExit("oglFramebuffer::bind","MISSING_ATTACHMENT_EXT");
      //  case GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT:
      //          oglExit("oglFramebuffer::bind","DUPLICATE_ATTACHMENT_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
                oglExit("oglFramebuffer::bind","DIMENSIONS_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
                oglExit("oglFramebuffer::bind","FORMATS_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
                oglExit("oglFramebuffer::bind","DRAW_BUFFER_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
                oglExit("oglFramebuffer::bind","READ_BUFFER_EXT");
        default:
                oglExit("oglFramebuffer::bind","GL_FRAMEBUFFER  logic error");
        };
}

/* Make the current texture "work right", with the usual filtering */
void tex_workright(void) {
	/* Texture state applies to the currently bound texture */
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,16);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_LINEAR);
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	//glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
}

#define glUniformTextures(prog) \
			glUniform1iARB(glGetUniformLocationARB(prog,"tex"),0);\
			glUniform1fARB(glGetUniformLocationARB(prog,"texscale"),texscale);\
			glUniform1fARB(glGetUniformLocationARB(prog,"texscalei"),texscalei);

/* Given plane coordinates (PL), output texture coordinates (TX) */
float TXfmPL(float x) {
	x*=texscale;
	return x/sqrt(1.0f+x*x)*0.5f+0.5f;
}
vec2 TXfmPL(const vec2 &p) {
	return vec2(TXfmPL(p.x),TXfmPL(p.y));
}
/* Given texture coordinate (TX), output plane coordinate (PL) */
float PLfmTX(float s) {
	float u=2.0*s-1.0;
	return texscalei*u/sqrt(1.0-u*u);
}
vec2 PLfmTX(const vec2 &s) {
	return vec2(PLfmTX(s.x),PLfmTX(s.y));
}

std::string GLSL_PLutil=STRINGIFY_GLSL(
	/* GLSL PLutil code */
	const float M_PI=3.14159265358979;
	const float M_1_PI=1.0/M_PI;
	const float EPS=1.0e-6;
	uniform float texscale;
	uniform float texscalei;

	/* Given plane coordinates (PL), output texture coordinates (TX) */
	float TXfmPL(float x) {
		x*=texscale;
		return x/sqrt(1.0+x*x)*0.5+0.5;
	}
	vec2 TXfmPL(vec2 p) {
		return vec2(TXfmPL(p.x),TXfmPL(p.y));
	}
	/* Given texture coordinate (TX), output plane coordinate (PL) */
	float PLfmTX(float s) {
		float u=2.0*s-1.0;
		return texscalei*u/sqrt(1.0-u*u);
	}
	vec2 PLfmTX(vec2 s) {
		return vec2(PLfmTX(s.x),PLfmTX(s.y));
	}
);

/* Compare TXfmPL expressions: */
	float texscale_tan=1.0;
	/* Given plane coordinates (PL), output texture coordinates (TX) */
	float TXfmPL_tan(float x) {
		x*=texscale_tan;
		return atan(x)*(1.0/M_PI)+0.5;
	}
	/* Given texture coordinate (TX), output plane coordinate (PL) */
	float PLfmTX_tan(float s) {
		float c=s*M_PI-(0.5*M_PI);
		float x=tan(c)/texscale_tan;
		return x; //if (s<0.5) return -x; else return +x;
	}

	/* Given plane coordinates (PL), output texture coordinates (TX) */
	vec2 TXfmPL_polar(vec2 p) {
		p*=texscale;
		float ir=0.5/sqrt(1+p.x*p.x+p.y*p.y);
		return ir*p+vec2(0.5);
		/* Final output r = ir*length(p)
			ir = 1/sqrt(k+length^2(p))
			output r = p/sqrt(k+p^2)
			r^2 (k+p^2) = p^2
			r^2 + (r^2-k)p^2 = 0
			p = sqrt(r^2/(k-r^2)), but then you lose an r from the existing scale
		*/
	}
	/* Given texture coordinate (TX), output plane coordinate (PL) */
	vec2 PLfmTX_polar(vec2 s) {
		vec2 t=2.0*s-vec2(1.0);
		float r=length(t);
		float det=1.0-r*r;
		//if (det<0.0) discard;
		return t*texscalei/sqrt(det);
	}

void TXfmPL_dump(void) {
	float ot=texscale;
	texscale=0.5;
	std::ofstream of("TXfmPL.txt");
	for (double pl=-3.0;pl<=3.0;pl+=1.0/32.0) {
		of<<pl<<"	"<<TXfmPL(pl)<<"	"<<TXfmPL_tan(pl)<<"	"<<TXfmPL_polar(vec2(pl,pl)).x<<"\n";
	}
	texscale=ot; texscalei=1.0/ot;
}


vec3 operator*(const mat4 &m,const vec3 &v) {
	return vec3(m*vec4(v,1));
}
void draw_texture(mat4 m,vec3 lo,vec3 hi)
{
	glBegin (GL_TRIANGLE_FAN);
        glTexCoord2f(0,0); glVertex3fv(m*vec3(lo.x,lo.y,lo.z)); 
        glTexCoord2f(1,0); glVertex3fv(m*vec3(hi.x,lo.y,lo.z)); 
        glTexCoord2f(1,1); glVertex3fv(m*vec3(hi.x,hi.y,lo.z)); 
        glTexCoord2f(0,1); glVertex3fv(m*vec3(lo.x,hi.y,lo.z)); 
        glEnd();
}


/************ IFS Map functions **************/
enum {max_xForm=10}; /* max GUI xForm editors */
xForm xfm_gui[max_xForm]; /* SILLY: this needs to be an array (not a vector) for GLUI to work... */
void copy_to_GUI(const std::vector<xForm> &xfm) {
	for (unsigned int m=0;m<max_xForm;m++)
	{
		if (m<xfm.size()) {
			xfm_gui[m]=xfm[m];
		}
		else {
			char tag[100]; sprintf(tag,"%d",m);
			xfm_gui[m]=xForm(tag);
		}
	}
	GLUI_Master.sync_live_all(); /* update GLUI copies, and redisplay */
}
std::vector<xForm> copy_from_GUI(void) {
	std::vector<xForm> xfm;
	for (int m=0;xfm_gui[m].weight>0.0 && m<max_xForm;m++)
		xfm.push_back(xfm_gui[m]);
	return xfm;
}

std::vector<xForm> xfm; /* used for rendering */

/* Return the color of IFS map m */
vec4 mapColor(int m) {
	float r=1.0,g=1.0,b=1.0;
	switch(m%3) {
	case 0: r-=colorful; break;
	case 1: g-=colorful; break;
	case 2: b-=colorful; break;
	}
	if (map_exposure<0.0) {
		std::swap(r,g); /* reddish, not greenish */
		std::swap(b,g); /* purplish, not greenish */
	}
	float e=fabs(map_exposure)*xfm[m].weight;
	//glColor3f(e*r,e*g,e*b);
	return vec4(e*r,e*g,e*b,1.0);
}



/* Run one step of the simulation for all the per-pixel maps for the current IFS */
void simulate_perpixel_maps(int level) 
{
	static std::vector<xForm> xfm_cached_px; /* used to detect changes to xfm array */
	static GLhandleARB prog=0; /* saved GLSL program for *all* per-pixel maps */
	if (prog==0 || !(xfm_cached_px==xfm)) 
	{/* Rebuild shaders */
		xfm_cached_px=xfm;
		/* Build up the unified inverse maps codebase */
		std::string inverseFns="", sumFn="";

		for (unsigned int m=0;m<xfm.size();m++) 
		if (xfm[m].hasInverse())
		{
			/* assert-check the transform */
			verify_inverse(&xfm[m],0);
			std::string tag=xfm[m].tag;

			inverseFns+=
			"/*------------------- w"+tag+" -----------------*/\n"
			"uniform vec4 color"+tag+";\n"
			"vec4 g"+tag+"(vec2 p) {\n"
			"	return exp2(texture2D(tex,TXfmPL(p))*20.0)-1.0;\n"
			"}\n"
			+xfm[m].getInverseCode();

			sumFn+=
			"	sum+=color"+tag+"*nonlinear_inverse"+tag+"(p);\n";
		}

	/* Compile the shaders */
		prog=makeProgramObject(STRINGIFY_GLSL(
		// GLSL Vertex shader (trivial)
		varying vec2 destcoords;
		void main(void) {
			destcoords = vec2(gl_Vertex);
			gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex; 
		}
		),verbose_shaders("per-pixel inverse",
		// GLSL Fragment shader
		"varying vec2 destcoords;\n"
		"uniform sampler2D tex;\n"
		+GLSL_PLutil+"\n"
		+inverseFns+
		"/* Combined inverse-sampling function */\n"
		"vec4 sum_inverses(vec2 p) {\n"
			"	vec4 sum=vec4(0.0);\n"
			+sumFn+"\n"
			"	return log2(sum+1.0)*(1.0/20.0);\n"
			"}\n"
		+STRINGIFY_GLSL(
		// GLSL Fragment shader main
		void main(void) {
        		gl_FragColor = sum_inverses(PLfmTX(destcoords));
			//gl_FragColor = vec4(destcoords.x,destcoords.y,0.5,0.5);  // debug in blue
			//gl_FragColor.a*=0.93; /* only be opaque where jacobi is crushing in */
		}
		)).c_str());
	} /* done building unified per-pixel maps codebase */

        glUseProgramObjectARB(prog);
	glUniformTextures(prog);

	/* Set uniform variables for each map */
	for (unsigned int m=0;m<xfm.size();m++) 
	if (xfm[m].hasInverse())
	{
		xfm[m].setUniforms(prog);
		glUniform4fvARB(glGetUniformLocationARB(prog,
			("color"+xfm[m].tag).c_str()),1,mapColor(m));
	}

	/* Draw screen-filling quad */
	draw_texture(mat4(1.0), vec3(0.0,0.0,0.5), vec3(1.0,1.0,0.5));
	glUseProgramObjectARB(0);
}

/* Run one step of the simulation for this per-vertex map of the current IFS */
void simulate_pervertex_map(int level,int m) 
{
	static std::vector<GLhandleARB> xfm_prog(xfm.size(),(GLhandleARB)0); /* cached compiled programs */
	static std::vector<xForm> xfm_cached_vx(xfm.size()); /* to detect changes in xfms */
	if (xfm_prog[m]==0 || !(xfm[m]==xfm_cached_vx[m])) 
	{ /* need to rebuild cached shader */
		xfm_cached_vx[m]=xfm[m];
		/* Build and compile geometry-rendering shader */
		xfm_prog[m]=makeProgramObject(
		    verbose_shaders("per-vertex maps",GLSL_PLutil
			+xfm[m].getCode()+
			"// GLSL Vertex shader:\n"
			"vec2 applyMap(vec2 t) {\n"
			"	return TXfmPL(applyMap"+xfm[m].tag+"(PLfmTX(t)));\n"
			"}\n"
			STRINGIFY_GLSL(
			varying vec4 color;
			varying vec2 texcoords;
        		void main(void) {
				texcoords=vec2(gl_Vertex);
				vec2 onscreen=applyMap(texcoords);
				color=getDensity(texcoords)*gl_Color;
				gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * vec4(onscreen,0,1); 
        		}
		)).c_str(),STRINGIFY_GLSL(
			// GLSL Fragment shader (trivial)
			varying vec4 color;
        		varying vec2 texcoords;
        		uniform sampler2D tex;
        		void main(void) {
        			gl_FragColor = color*texture2D(tex,texcoords);
				//gl_FragColor = vec4(texcoords.x,texcoords.y,0.5,0.5);  // debug in blue
        		}
		));
	} /* end building vertex shader */

        glUseProgramObjectARB(xfm_prog[m]);
	glUniformTextures(xfm_prog[m]);
	xfm[m].setUniforms(xfm_prog[m]);
	glColor4fv(mapColor(m));

	int res=disc_compute;
	//while ((1<<(level-1))>res && res>4) res/=2; /* use less geometry for low multigrid levels */

	double dy=1.0/res, dx=1.0/res;
	glBegin(GL_TRIANGLE_STRIP);
	for (double y=0;y<1.0;y+=dy) {
		for (double x=0;x<=1.0;x+=dx) {
			vec2 t=vec2(x,y);
			glVertex2fv(t);
			if (x==0) glVertex2fv(t);
			t=vec2(x,y+dy);
			glVertex2fv(t);
			if (x+dx>1.0) glVertex2fv(t); /* last on row */
		}
	}
	glEnd();

        glUseProgramObjectARB(0);
}

/* Run a multigrid simulation of the current IFS, and return the finished texture. */
GLuint simulate_IFS(void) 
{
	texscalei=1.0/texscale;
	
	glMatrixMode(GL_PROJECTION); /* clear out projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* Framebuffer output textures 
		We really need multigrid to get decent convergence at framerate.
	*/
	static GLuint texLevels[nlevel+1][2]={{0}}; /* master copy */
	GLuint frameSrc=0, frameDest; /* working copy, pingponged buffers */
	if (texLevels[0][0]==0) 
	{  /* build a texture for every possible multigrid level */
		glGenTextures(2*(nlevel+1),&texLevels[0][0]); /* make texture handles */
		for (int level=0;level<=nlevel;level++)
		for (int i=0;i<2;i++) {
			glBindTexture(GL_TEXTURE_2D,texLevels[level][i]);
			glTexImage2D(GL_TEXTURE_2D,0,
				GL_RGBA8,1<<level,1<<level, /* data format and size (pixels) */
				0,GL_LUMINANCE,GL_FLOAT,0);
			glGenerateMipmapEXT(GL_TEXTURE_2D); /*<- needed on many cards! */
			tex_workright();
		}
	}
	
	/* Framebuffer object.  Only one needed, with attachments swapped around. */
	static GLuint fbo=0;
	if (fbo==0) { /* create up framebuffer object--set it up below */
		glGenFramebuffersEXT(1,&fbo);
	}
	
	/* Render into our framebuffer object */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE); /* pure additive alpha */
	glDisable(GL_CULL_FACE); /* some maps produce backwards triangles */

	glScalef(1.0,1.0,0.01); /* scale down Z axis, to avoid z-clipping */

	glScalef(2.0,2.0,1.0); /* center X and Y on 0.5,0.5, like texture coordinates */
	glTranslatef(-0.5,-0.5,0.0);
	
	/* Loop over multigrid levels */
	for (int level=firstlevel;level<=lastlevel;level++)
	{
		/* When starting a new multigrid level, just swap the destination,
		  using the old source texture.  Texcoords will automatically interpolate! */
		frameDest=texLevels[level][0];
		if (level==firstlevel) { /* need a source too */
			frameSrc=texLevels[level][1];
		}
		int w=1<<level, h=1<<level; /* pixels on this level (dest) */
	
	/* Iteratively draw the IFS maps */
	for (int depth=0;depth<iterations;depth++)
	{
		// Make sure we're rendering into frameTex[1]
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
			GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,
			frameDest, 0);
		/* Make sure the framebuffer's happy */
		GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
			bad_framebuffer(status);
		glDisable(GL_DEPTH_TEST); /* or else attach to GL_DEPTH_ATTACHMENT_EXT! */
		glViewport(0,0,w,h);
		glScissor(0,0,w,h);
		
		// Clear the screen
		if (depth==0 && level==firstlevel) 
		{ /* starting a new computation-- total reset */
			float g=screen_initval; /* initialize color to this */
			int drawWhat=1;
			if (drawWhat!=0) g=0.0; /* black background */
			
			glClearColor(g,g,g,1.0f); 
			glClear(GL_COLOR_BUFFER_BIT);
			
			glDisable(GL_TEXTURE_2D);
			glPushMatrix();
			glTranslatef(0.5,0.5,0.0);
			float size=screen_initscale*(TXfmPL(1.0)-TXfmPL(0.0)); // estimate unit-plane size in texture
			glColor3f(screen_initval,screen_initval,screen_initval);
			switch (drawWhat) {
			case 1: glutSolidCube(size); break;
			case 2: glutSolidSphere(size,10,20); break;
			};
			glPopMatrix();
			glColor3f(1,1,1);
			
			glFinish();
			
			// Build mipmaps of our freshly-rendered data
			glBindTexture(GL_TEXTURE_2D,frameDest);
			glEnable(GL_TEXTURE_2D);
			glGenerateMipmapEXT(GL_TEXTURE_2D); 
		
			std::swap(frameSrc,frameDest);
			continue; /* go around again to draw maps */
		} 
		/* else not the first frame--we'll erase during the per-pixel inverse */
		
		// Draw our IFS estimate frameSrc, transformed by each map
		glBindTexture(GL_TEXTURE_2D,frameSrc);
		glEnable(GL_TEXTURE_2D);
		
/************ Per-pixel inverse functions *************/
		glDisable(GL_BLEND); /* avoids erase--hard-overwrite previous texture */
		if (!force_vertexonly) {
			simulate_perpixel_maps(level);
		} else { /* vertex-only mode--still have to erase screen */
			glClearColor(0,0,0,1.0f); glClear(GL_COLOR_BUFFER_BIT);
		}
		
/************ If no inverse function, approximate the inverse using geometry *********/
		glEnable(GL_BLEND); /* need to combine per-pixel with geometry (FIXME: log textures?) */
		for (unsigned int m=0;m<xfm.size();m++) 
		  if (!xfm[m].hasInverse())
			simulate_pervertex_map(level,m);
		/* end geometric approximate inverse */
		
		// Build mipmaps of our freshly-rendered data
		glBindTexture(GL_TEXTURE_2D,frameDest);
		glGenerateMipmapEXT(GL_TEXTURE_2D); 
		
		// Swap textures zero and one
		std::swap(frameSrc,frameDest);
		if (depth==0) 
		{ /* done with first pass through a new multigrid level--now update other buffer */
			frameDest=texLevels[level][1];
		}
	} /* end multipass depth loop */
	
	if (level==lastlevel && save_image) { /* dump whole texture */
		char fname[100]; sprintf(fname,"tex%03d.ppm",save_count);
		oglSavePPM(0,0,w,h,fname);
	}
} /* end multigrid loop */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
	
	return frameSrc; /* we flipped src around to dest */
}

/* Bind in an offscreen framebuffer object with this size */
void bind_offscreen_framebuffer(int w,int h) {
	int fb=0;
	static GLuint texLevels[nlevel+1][2]={{0}}; /* master copy */
	GLuint frameDest=0; /* texture */
	glGenTextures(1,&frameDest); /* make texture handles */
	glBindTexture(GL_TEXTURE_2D,frameDest);
	glTexImage2D(GL_TEXTURE_2D,0,
		GL_RGBA8,w,h,0,GL_RGBA,GL_FLOAT,0);
	glGenerateMipmapEXT(GL_TEXTURE_2D); /*<- needed on many cards! */
	
	/* Framebuffer object.  Only one needed, with attachments swapped around. */
	static GLuint fbo=0;
	if (fbo==0) { /* create up framebuffer object--set it up below */
		glGenFramebuffersEXT(1,&fbo);
	}
	/* Render into our framebuffer object */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,
		frameDest, 0);
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
		bad_framebuffer(status);
}

void my_display(void) {
	glFinish(); int last_msec=glutGet(GLUT_ELAPSED_TIME);
	
	// Switching presets:
	//static std::vector<xForm> xfm_src; /* xfm, before animation */
	static int fractal_predefined_old=-1;
	if (fractal_predefined!=fractal_predefined_old) 
	{ /* New preset */
		fractal_predefined_old=fractal_predefined;
		copy_to_GUI(make_xfms());
	}
	else { /* Same preset, but maybe GUI modifications */
		
	}
	
	// Animation support:
	static float animate_start=-1; // milliseconds at start of animation
	static bool was_anim=false;
	if (animate_maps) {
		xfm=copy_from_GUI();
		if (!was_anim) animate_start=last_msec;
		
		if (save_movie) { /* Animate at fixed 30fps for output */
			animate_start=0;
			last_msec=(int)(save_count*(1.0/30)*1000);
		}
		
		was_anim=true;
		glutPostRedisplay(); /* we're now time-varying */
		for (unsigned int m=0;m<xfm.size();m++) 
		{ /* rotation rate (rads/msec) is slightly different for different maps */
			double rate=0.001*0.2*(1+0.1234*m);
			double angle=rate*(last_msec-animate_start);
			//printf("Angle=%.2f rads\n",angle);
			float c=cos(angle), s=sin(angle);
			vec2  tx=xfm[m].x*c - xfm[m].y*s;
			xfm[m].y=xfm[m].x*s + xfm[m].y*c;
			xfm[m].x=tx;
		}
	} else { /* done animating--copy *back* the modified maps */
		if (was_anim) {
			copy_to_GUI(xfm);
		} else { /* normal non-animating frame */
			xfm=copy_from_GUI();
		}
		was_anim=false;
	}
	
	GLuint texIFS=simulate_IFS();
	
	int render_w=glutGet(GLUT_WINDOW_WIDTH), render_h=glutGet(GLUT_WINDOW_HEIGHT);
	if (save_image) { /* render to offscreen high-res image */
		render_w=1<<lastlevel; render_h=1<<lastlevel;
		bind_offscreen_framebuffer(render_w,render_h);
	}
	
	glViewport(0,0,render_w,render_h);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

/* Do normal fullscreen rendering */
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		render_w/(float)render_h, /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	/* Draw the current output texture */
	glBindTexture(GL_TEXTURE_2D,texIFS);
	glEnable(GL_TEXTURE_2D); 
	glColor4f(1,1,1,1);
	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	/* Do final plane-from-texture conversion exactly, in pixel shader */
	static GLhandleARB plane_prog=makeProgramObject(STRINGIFY_GLSL(
		// GLSL Vertex shader:
		varying vec3 worldCoords;
		void main(void) {
			worldCoords = vec3( gl_ModelViewMatrix * gl_Vertex );
			gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex; 
		}
	),verbose_shaders("final output",GLSL_PLutil+STRINGIFY_GLSL(
		// GLSL Fragment shader
		varying vec3 worldCoords;
		uniform sampler2D tex;
		uniform float colormode;

		void main(void) {
			/* flip display Y coordinate, to match Apophysis rendering... */
			vec2 texcoords=TXfmPL(vec2(worldCoords.x,-worldCoords.y));
			//texcoords=vec2(worldCoords); // for debugging texture generation
			vec4 t = texture2D(tex,texcoords);
			gl_FragColor = t*colormode + (1.0-t)*(1.0-colormode);
			gl_FragColor.a=t.a; /* copy alpha unmodified */
		}
	)).c_str());
        glUseProgramObjectARB(plane_prog);
	glUniformTextures(plane_prog);
	glFastUniform1f(plane_prog,"colormode",map_exposure<0?0.0:1.0);
	float gs=100; /* ground plane dimensions: over 1000 gives roundoff */
	draw_texture(mat4(1.0), vec3(-gs,-gs,0.5), vec3(gs,gs,0.5));
        glUseProgramObjectARB(0);
	
	
	if (save_image || save_movie) { /* dump non-distored screen */
		char fname[100]; sprintf(fname,"image%03d.ppm",save_count++);
		oglSavePPM(0,0,render_w,render_h,fname);
		save_image=0;
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
		
		/* Typical usage:
			glutInitWindowSize(1280,720);
			Run to capture frames
			~olawlor/dev/osl/bin/img2mov out.avi 'image%03d.ppm' 0.033333
			Add audio and re-encode with kdenlive, iMovie, etc.
		*/
	}

	glutSwapBuffers(); /* copy back buffer to front */
	
/* Statistics */
	glFinish();int cur_msec=glutGet(GLUT_ELAPSED_TIME); 
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
}

/******* Function Performance Profiling ***********/
/**
  Return the current time in seconds (since something or other).
*/
#if defined(WIN32)
#  include <sys/timeb.h>
#  define time_in_seconds_granularity 0.1 /* seconds */
double time_in_seconds(void) { /* This seems to give terrible resolution (60ms!) */
        struct _timeb t;
        _ftime(&t);
        return t.millitm*1.0e-3+t.time*1.0;
}
#else /* UNIX or other system */
#  include <sys/time.h> //For gettimeofday time implementation
#  define time_in_seconds_granularity 0.05 /* seconds */
double time_in_seconds(void) {
	struct timeval tv;
	gettimeofday(&tv,NULL);
	return tv.tv_usec*1.0e-6+tv.tv_sec*1.0;
}
#endif


void print_sim_time(const char *name,int value, float scale) {
	enum {nruns=5};
	double elapsed[nruns];
	for (int run=0;run<nruns;run++) {
		double start=time_in_seconds();
		
		simulate_IFS();
		glFinish();
		
		elapsed[run]=time_in_seconds()-start;
	}
	std::sort(&elapsed[0],&elapsed[nruns]);
	double time=elapsed[nruns/2];
	char buf[1000];
	sprintf(buf,name,value); 
	printf("%-30s	%.3fms\n",buf,time*scale*1.0e3); fflush(stdout);
}

int main(int argc,char *argv[]) {
	verify_inverses();

	glutInitWindowSize(1280,720);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	int win=glutCreateWindow("Nonlinear IFS on the GPU");
	glewInit();
	
	int argi=1;
	if (argc>argi && argv[argi]==std::string("-vertex")) { 
		argi++;
		force_vertexonly=1;
	}
	
	if (argc>argi && argv[argi]==std::string("-bench")) { argi++;
		print_sim_time("default mode (fouling)",0,1.0);
		
		for (int mgiter=3;mgiter<=10;mgiter+=7)
		for (int output=9;output<=nlevel;output++) {
			char buf[100];
			int spanlevels=7;
			iterations=mgiter;
			printf("--output %d pixels--\n",1<<output);
			firstlevel=output-spanlevels; lastlevel=output;
			sprintf(buf,"multigrid %d-%d x%d",firstlevel,lastlevel,iterations); print_sim_time(buf,0,1.0);
			
			iterations=mgiter*(spanlevels+1); firstlevel=lastlevel;
			sprintf(buf,"fixed %d x%d",lastlevel,iterations); print_sim_time(buf,0,1.0);
		}
		
		firstlevel=5; lastlevel=10;
		for (iterations=1;iterations<=10;iterations++)
			print_sim_time("multigrid 5-10 x%d",iterations, 1.0);
		firstlevel=5; lastlevel=11;
		for (iterations=1;iterations<=10;iterations++)
			print_sim_time("multigrid 5-11 x%d",iterations, 1.0);
		firstlevel=5; lastlevel=12;
		for (iterations=1;iterations<=10;iterations++)
			print_sim_time("multigrid 5-12 x%d",iterations, 1.0);
		
		for (int level=0;level<=nlevel;level++) {
			firstlevel=lastlevel=level; iterations=50;
			print_sim_time("pixel%d/iteration",1<<level, 1.0/iterations);
			
			force_vertexonly=1;
			disc_compute=10;
			print_sim_time("10v %d/iteration",1<<level, 1.0/iterations);
			disc_compute=100;
			print_sim_time("100v %d/iteration",1<<level, 1.0/iterations);
			disc_compute=1000;
			print_sim_time("1000v %d/iteration",1<<level, 1.0/iterations);
			
			force_vertexonly=0;
		}
		return 0;
	}
	
	TXfmPL_dump();
	
	glutDisplayFunc(my_display);
	oglCameraInit();
	camera_orient.x=vec3(1,0,0);
	camera_orient.y=vec3(0,1,0);
	camera_orient.z=vec3(0,0,1);
	//camera=vec3(0.5,0.5,+1);
	camera=vec3(0.0,0.0,+3);
	
// Build the user interface
	GLUI *mainglui = GLUI_Master.create_glui_subwindow(win, GLUI_SUBWINDOW_LEFT);
	mainglui->set_main_gfx_window( win );
	GLUI_Node *glui=mainglui;
	
	(new GLUI_Spinner(glui,"Preset",&fractal_predefined))->set_int_limits(0,6);
	new GLUI_Checkbox(glui,"Animate",&animate_maps);
	new GLUI_Checkbox(glui,"Save Image",&save_image);
	new GLUI_Checkbox(glui,"Save Movie",&save_movie);
	
	glui=new GLUI_Rollout(mainglui,"Rendering");
	//new GLUI_Spinner(glui,"Bias",&map_bias);
	new GLUI_Spinner(glui,"exposure",&map_exposure);
	new GLUI_Spinner(glui,"color",&colorful);
	new GLUI_Spinner(glui,"texscale",&texscale);
	new GLUI_Spinner(glui,"init grey",&screen_initval);
	new GLUI_Spinner(glui,"init size",&screen_initscale);
	//new GLUI_Column(glui);
	(new GLUI_Spinner(glui,"firstlevel:2^",&firstlevel))->set_int_limits(0,nlevel);
	(new GLUI_Spinner(glui,"lastlevel:2^",&lastlevel))->set_int_limits(0,nlevel);
	new GLUI_Spinner(glui,"iter/level",&iterations);	
	//new GLUI_Column(glui);
	new GLUI_Checkbox(glui,"Per-vertex",&force_vertexonly);
	(new GLUI_Spinner(glui,"vertices",&disc_compute))->set_int_limits(1,2000);
	//new GLUI_Column(glui);
	
	xfm=make_xfms();
	for (unsigned int m=0;m<max_xForm;m++) {
		char mapname[100]; sprintf(mapname,"Map %d",m);
		glui=new GLUI_Rollout(mainglui,mapname);
		//new GLUI_Rotation(glui,"Mat1",&mat1[0][0]);
		(new GLUI_Spinner(glui,"Variation",&xfm_gui[m].var))
			->set_int_limits(0,sizeof(variation_list)/sizeof(variation_list[0])-2);
		new GLUI_Spinner(glui,"Weight",&xfm_gui[m].weight);
		new GLUI_Spinner(glui,"Scale",&xfm_gui[m].wvar);
		new GLUI_Spinner(glui,"X.X",&xfm_gui[m].x.x);
		new GLUI_Spinner(glui,"X.Y",&xfm_gui[m].x.y);
		new GLUI_Spinner(glui,"Y.X",&xfm_gui[m].y.x);
		new GLUI_Spinner(glui,"Y.Y",&xfm_gui[m].y.y);
		new GLUI_Spinner(glui,"O.X",&xfm_gui[m].o.x);
		new GLUI_Spinner(glui,"O.Y",&xfm_gui[m].o.y);
		//new GLUI_Column(glui);
	}
	
	glutMainLoop(); /* processes events */
	return 0;
}
